
RELEASE_VERSION='0.0.393010107'
BUILD_DATE='Thu Oct 21 17:58:23 UTC 2021'
BUILD_MACHINE_INFO='Linux runner-mvj-ljo-project-28359933-concurrent-0xnhr6 5.4.129-62.227.amzn2.x86_64 #1 SMP Wed Jul 7 00:08:43 UTC 2021 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='1703331830'
GIT_COMMIT_SHA='e294b016729996a317363efc069024036b876312'
GIT_BRANCH='main'
